"""
This module includes the Proton VPN GTK application for Linux.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from importlib.metadata import version, PackageNotFoundError
import gi

try:
    __version__ = version("proton-vpn-gtk-app")
except PackageNotFoundError:
    __version__ = "development"

gi.require_version("Gtk", "4.0")
gi.require_version("Notify", "0.7")

from gi.repository import Gtk  # pylint: disable=C0413 # noqa: E402

from proton.vpn import logging  # pylint: disable=C0413 # noqa: E402


logging.config(filename="vpn-app")

__all__ = [Gtk]



"""
This module defines the port forwarding widget.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional
from gi.repository import Gdk, GObject
from proton.vpn.app.gtk import Gtk
from proton.vpn.connection import states

from proton.vpn.app.gtk.widgets.main.notifications import Notifications


class PortForwardRevealer(Gtk.Revealer):  # pylint: disable=too-few-public-methods
    """The container that has all PF widgets and reveals on demand."""
    def __init__(self,
                 notifications: Notifications,
                 port_forward_widget: Optional["PortForwardWidget"] = None):
        super().__init__()
        self._port_forward_widget = \
            port_forward_widget or PortForwardWidget(notifications)
        self.set_child(self._port_forward_widget)
        self._port_forward_widget.connect(
            "update-visibility", self._on_update_port_forwarding_visibility
        )

    def on_new_state(self, connection_state: states.State):
        """Proxy method that relays connection state changes to PF widget."""
        self._port_forward_widget.on_new_state(connection_state)

    def _on_update_port_forwarding_visibility(self, _: PortForwardWidget, display_child: bool):
        self.set_reveal_child(display_child)


class PortForwardWidget(Gtk.Button):
    """Widgets handles the display and interactivity to copy por to clipboard."""
    ACTIVE_PORT_LABEL = "Active port:"
    TOOLTIP_LABEL = "Copy port number"

    def __init__(
            self, notifications: Notifications, clipboard: Optional[Gdk.Clipboard] = None,
            forwarded_port: Optional[int] = None
    ):
        super().__init__()
        self.set_name("port-forwarding-widget")
        self._notifications = notifications
        self._clipboard = (
            clipboard or Gdk.Display.get_default().get_clipboard()
        )
        self._current_forwarded_port = forwarded_port
        self._build_ui()

        cursor = Gdk.Cursor.new_from_name("pointer", None)
        self.set_cursor(cursor)

    @GObject.Signal(name="update-visibility", arg_types=(bool,))
    def update_visibility(self, display_child: bool):
        """
        Signal emitted when the UI should be hidden or not.
        :param display_child: whether PF should be displayed or not.
        """

    def _build_ui(self):
        self.set_halign(Gtk.Align.CENTER)
        self.set_property("margin-top", 10)
        self.set_tooltip_text(self.TOOLTIP_LABEL)

        # Create the label that will contain the static text
        active_port_label = Gtk.Label(label=self.ACTIVE_PORT_LABEL)
        active_port_label.add_css_class("dim-label")

        # Create the label that will contain the active port, so that it can
        # be easily copied to clipboard
        self._port_forward_label: Gtk.Label = Gtk.Label(label="")
        self._port_forward_label.add_css_class("dim-label")

        # Create the copy icon
        copy_port_to_clipboard_image = Gtk.Image.new_from_icon_name("edit-copy")
        copy_port_to_clipboard_image.add_css_class("dim-label")

        # Create a label box that will contain the port info. This will us to give
        # more breathing room between text and icon.
        label_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        label_box.append(active_port_label)
        label_box.append(self._port_forward_label)
        label_box.set_spacing(3)

        # This box will contain both the label and the icon,
        # for easier styling manipulation.
        content_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        content_box.append(label_box)
        content_box.append(copy_port_to_clipboard_image)
        content_box.set_spacing(10)

        self.set_child(content_box)

        # set up click to copy port to clipboard
        self.connect("clicked", self._on_button_press)

    def on_new_state(self, connection_state: states.State):
        """Receives new connection state and emits a signal
        of whether it should be hidden or not."""
        self._update_visibility(
            connection_state.forwarded_port,
            reveal_child=isinstance(connection_state, states.Connected)
        )

        top_level = self._port_forward_label.get_root()
        is_focus = True

        # We need to check that the top level widget is indeed the main window
        # because when shutting down the app this widget could process events
        # after the main window is already closed.
        if isinstance(top_level, Gtk.Window):
            is_focus = top_level.is_active()

        new_port = connection_state.forwarded_port
        if new_port and (new_port != self._current_forwarded_port):
            if not is_focus:
                self._notifications.show_gnome_notification(
                    title="Port forwarding",
                    description=f"Active port is {new_port}"
                )

        self._current_forwarded_port = new_port

    def _update_visibility(self, forwarded_port: Optional[int], reveal_child: bool):
        if forwarded_port is None:
            self.emit("update-visibility", False)
            return

        self.emit("update-visibility", reveal_child)
        self.set_port_forward_label(forwarded_port)

    def _on_button_press(
        self, _: "PortForwardWidget"
    ):
        port_to_be_copied_to_clipboard = self._port_forward_label.get_label()
        value = GObject.Value(GObject.TYPE_STRING, port_to_be_copied_to_clipboard)
        self._clipboard.set(value)

    def set_port_forward_label(self, new_port: int):
        """Helper method to set port forward label."""
        self._port_forward_label.set_label(str(new_port))



"""
This module defines the main application window.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from os import environ
from typing import Optional

from gi.repository import Gtk

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.main.main_widget import MainWidget
from proton.vpn.app.gtk.widgets.headerbar.headerbar import HeaderBar
from proton.vpn.app.gtk.widgets.main.notification_bar import NotificationBar
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.main.pull_notifications.nps_survey_modal import \
    NPSSurvey


class MainWindow(Gtk.ApplicationWindow):
    """Main window."""

    WIDTH = 450
    HEIGHT = 700

    # pylint: disable=too-many-arguments
    def __init__(
            self, application: Gtk.Application,
            controller: Controller,
            notifications: Optional[Notifications] = None,
            header_bar: Optional[HeaderBar] = None,
            main_widget: Optional[MainWidget] = None,
            overlay_widget: Optional[OverlayWidget] = None
    ):
        super().__init__(application=application)
        self._application = application
        self.get_settings().props.gtk_application_prefer_dark_theme = True
        self._controller = controller
        self._close_window_handler_id: Optional[int] = None
        self._shortcut_controller: Optional[Gtk.ShortcutController] = None

        self._configure_window()

        self._overlay_widget = overlay_widget or OverlayWidget()

        notifications = notifications or Notifications(
            main_window=self, notification_bar=NotificationBar()
        )

        self.header_bar = header_bar or HeaderBar(
            controller=controller,
            main_window=self,
            overlay_widget=self._overlay_widget
        )
        self.set_titlebar(self.header_bar)
        self.set_title("Proton VPN")

        self.main_widget = main_widget or MainWidget(
            controller=controller,
            main_window=self,
            notifications=notifications,
            overlay_widget=self._overlay_widget
        )
        self.set_child(self.main_widget)

        self.connect("show", self._display_pending_widget)

        self.main_widget.set_visible(True)

    @property
    def application(self) -> Gtk.Application:
        """Returns Gtk.Application object which contains references to windows,
        tray indicator and other settings."""
        return self._application

    def add_keyboard_shortcut(self, target_widget: Gtk.Widget, target_signal: str, shortcut: str):
        """
        Adds a keyboard shortcut so that when pressed it causes the target signal
        to be triggered on the target widget.

        :param target_widget: The widget the keyboard shortcut will trigger the signal on.
        :param target_signal: The signal the keyboard shortcut will trigger on the target widget.
        :param shortcut: The keyboard shortcut string (e.g. "<Ctrl>q")
        """
        def callback(*_):
            target_widget.emit(target_signal)
            return True

        shortcut_trigger = Gtk.ShortcutTrigger.parse_string(shortcut)
        if shortcut_trigger is None:
            raise ValueError(f"Invalid shortcut: {shortcut}")

        shortcut_action = Gtk.CallbackAction.new(callback)
        shortcut_obj = Gtk.Shortcut.new(shortcut_trigger, shortcut_action)

        if self._shortcut_controller is None:
            self._shortcut_controller = Gtk.ShortcutController()
            self._shortcut_controller.set_scope(Gtk.ShortcutScope.LOCAL)
            self.add_controller(self._shortcut_controller)

        self._shortcut_controller.add_shortcut(shortcut_obj)

    def _configure_window(self):
        """
        Handle delete-event, set window resize restrictions...
        """
        self.set_name("main-window")
        self.set_resizable(False)
        self.set_size_request(MainWindow.WIDTH, MainWindow.HEIGHT)

    def configure_close_button_behaviour(self, tray_indicator_enabled: bool):
        """Configures the behaviour of the button to close the window
        (the x button), depending on if the tray indicator is used or not."""
        if tray_indicator_enabled:
            self._close_window_handler_id = self.configure_close_button_to_hide_window()
        else:
            self._close_window_handler_id = self.configure_close_button_to_trigger_quit_menu_entry()

    def configure_close_button_to_hide_window(self):
        """Configures the x (close window) button so that when clicked,
        the window is hidden instead closed."""
        def on_close_button_clicked_then_hide_window(*_) -> bool:
            """
            Instead of letting the window x button close the app, therefore
            quitting the app, the action is delegated to the Exit entry in
            the menu bar widget.
            """
            self.set_visible(False)

            # Returning True when handling the close-request stops other handlers
            # from being invoked for this event, therefore preventing the default
            # behaviour:
            # https://docs.gtk.org/gtk4/signal.Window.close-request.html
            return True

        # Handle the event emitted when the user tries to close the window.
        return self.connect(
            "close-request",
            on_close_button_clicked_then_hide_window
        )

    def quit(self):
        """Closes the main window, which quits the app."""
        if self._close_window_handler_id:
            self.disconnect(self._close_window_handler_id)

        self.close()

    def configure_close_button_to_trigger_quit_menu_entry(self):
        """Configures the x (close window) button so that when clicked,
        the Exit menu entry is triggered instead."""
        def on_close_button_clicked_then_click_quit_menu_entry(*_) -> bool:
            """
            Instead of letting the x button close the app, therefore
            quitting the app, the action is delegated to the Exit entry in
            the menu bar widget, which may request confirmation to the user.
            """
            self.header_bar.menu.quit_button_click()

            # Returning True when handling the close-request stops other handlers
            # from being invoked for this event, therefore preventing the default
            # behaviour:
            # https://docs.gtk.org/gtk4/signal.Window.close-request.html
            return True

        # Handle the event emitted when the user tries to close the window.
        return self.connect(
            "close-request",
            on_close_button_clicked_then_click_quit_menu_entry
        )

    def _display_pending_widget(self, _):
        if "PROTON_VPN_FEATURE_FLAG_NPS" not in environ:
            return

        if self.is_visible():
            pending_attention_window = NPSSurvey(
                self._controller,
                submit_handler=lambda score, text: logging.info("NPS: %i %s", score, text),
                dismiss_handler=lambda: logging.info("NPS dismissed")
            )
            pending_attention_window.set_transient_for(self)
            pending_attention_window.show()



"""
User session monitoring.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Callable, Optional
import dbus
from dbus import SystemBus
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)


BUS_NAME = "org.freedesktop.login1"
SEAT_AUTO_PATH = "/org/freedesktop/login1/seat/auto"
SESSION_INTERFACE = "org.freedesktop.login1.Session"
SEAT_INTERFACE = "org.freedesktop.login1.Seat"
PROPERTIES_INTERFACE = "org.freedesktop.DBus.Properties"
UNLOCK_SIGNAL = "Unlock"


class SessionMonitor:
    """
    After being enabled, it calls the callback set on the
    session_unlocked_callback attribute whenever the user session was unlocked.

    Attributes:
        session_unlocked_callback: callable that will be called when the user
        session is unlocked.
    """
    def __init__(self, bus: SystemBus = None, session_object_path: Optional[str] = None):
        self._bus = bus
        self._session_object_path = session_object_path
        self._signal_receiver: Optional[object] = None
        self.session_unlocked_callback: Optional[Callable] = None

    def enable(self):
        """Enables user session monitoring."""
        if not callable(self.session_unlocked_callback):
            raise RuntimeError("Callback was not set")

        if not self._bus:
            self._bus = SystemBus()

        if not self._session_object_path:
            self._setup()

        self._signal_receiver = self._bus.add_signal_receiver(
            handler_function=self.session_unlocked_callback,
            signal_name=UNLOCK_SIGNAL,
            dbus_interface=SESSION_INTERFACE,
            bus_name=BUS_NAME,
            path=self._session_object_path,
        )

    def disable(self):
        """Disables user session monitoring"""
        if self._signal_receiver:
            self._signal_receiver.remove()
            self._signal_receiver = None

    @property
    def is_session_unlocked(self):
        """Returns True if the user session is unlocked or False otherwise."""
        if not self._session_object_path:
            self._setup()

        active_session = self._bus.get_object(BUS_NAME, self._session_object_path)
        active_session_properties = dbus.Interface(active_session, PROPERTIES_INTERFACE)
        return not bool(active_session_properties.Get(SESSION_INTERFACE, "LockedHint"))

    def _setup(self):
        seat_auto_proxy = self._bus.get_object(
            BUS_NAME,
            SEAT_AUTO_PATH
        )

        seat_auto_properties_proxy = dbus.Interface(
            seat_auto_proxy,
            PROPERTIES_INTERFACE
        )

        # There should always be session for a seat. If there is no seat then
        # it means that the user is not directly controlling the machine,
        # but rather controlloing it via ssh or some other indirect
        # type of control.
        seat_properties = seat_auto_properties_proxy.GetAll(SEAT_INTERFACE)
        active_sessions = seat_properties.get("ActiveSession", [])

        if not active_sessions:
            raise RuntimeError("There are no active sessions for this seat")

        _session_id, self._session_object_path = active_sessions

    def set_signal_receiver(self, new_object: Optional[object]):
        """Sets signal receiver.
        This is mainly used for testing purposes.
        """
        self._signal_receiver = new_object



"""
This module defines the country widgets.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass

from typing import List, Tuple, Set, Optional, Callable
from gi.repository import GLib, GObject

from proton.vpn.app.gtk.utils.accessibility import add_accessibility
from proton.vpn.app.gtk.utils.search import normalize
from proton.vpn.connection.enum import ConnectionStateEnum
from proton.vpn.session.servers import Country
from proton.vpn import logging
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import \
    SmartRoutingIcon, P2PIcon, TORIcon, UnderMaintenanceIcon
from proton.vpn.app.gtk.widgets.vpn.serverlist.server import ServerRow
from proton.vpn.session.servers import LogicalServer
from proton.vpn.session.servers import ServerFeatureEnum

logger = logging.getLogger(__name__)


@dataclass
class CountryAnalysis:
    """Contains a summary of the state all the servers in a given country."""
    country_connection_state: ConnectionStateEnum
    smart_routing_country: bool
    under_maintenance: bool
    is_free_country: bool
    country_features: Set[ServerFeatureEnum]


def _analyze_servers(ordered_servers: List[LogicalServer],
                     connected_server_id: Optional[str] = None) -> CountryAnalysis:
    """
    Iterates over the ordered list of servers and extracts information
    to be displayed for the country.
    """
    # Properties initialized after analysing the servers.
    is_free_country = None

    country_features = set()

    # The country is set under maintenance until the opposite is proven.
    under_maintenance = True

    # The country connection state is set as disconnected until the opposite is proven.
    country_connection_state = ConnectionStateEnum.DISCONNECTED

    # Smart routing is assumed to be used until the opposite is proven.
    smart_routing_country = True

    for server in ordered_servers:
        country_features.update(server.features)

        is_free_country = is_free_country or server.tier == 0

        # The country is under maintenance if (1) that was the case up until now and
        # (2) the current server is also under maintenance (i.e. is not enabled).
        under_maintenance = (under_maintenance and not server.enabled)

        # A country is flagged as a "Smart routing" location if *all* servers are
        # actually physically located in a neighboring country.
        smart_routing_country = \
            (smart_routing_country and server.host_country is not None)

        # If we are currently connected to a server then set its row state to "connected".
        if connected_server_id == server.id:
            country_connection_state = ConnectionStateEnum.CONNECTED

    return CountryAnalysis(
        country_connection_state,
        smart_routing_country,
        under_maintenance,
        is_free_country,
        country_features
    )


class CountryHeader(Gtk.Box):  # pylint: disable=too-many-instance-attributes
    """Header with the country name shown at the beginning of each CountryRow."""
    # pylint: disable=too-many-arguments
    def __init__(
            self,
            country: Country,
            under_maintenance: bool,
            upgrade_required: bool,
            server_features: Set[ServerFeatureEnum],
            smart_routing: bool,
            connection_state: ConnectionStateEnum,
            controller: Controller,
            show_country_servers: bool = False
    ):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL)
        self._country = country
        self._under_maintenance = under_maintenance
        self._upgrade_required = upgrade_required
        self._server_features = server_features
        self._smart_routing = smart_routing
        self._controller = controller

        self._country_name_label: Optional[Gtk.Label] = None
        self._under_maintenance_icon = None
        self._country_details = None

        self._connect_button: Optional[Gtk.Button] = None
        self._connect_button_handler_id: Optional[int] = None
        self._toggle_button: Gtk.Button
        self._toggle_button_handler_id: Optional[int] = None

        self._collapsed_img = Gtk.Image.new_from_icon_name("pan-down-symbolic")
        self._expanded_img = Gtk.Image.new_from_icon_name("pan-up-symbolic")

        self._build_ui(connection_state)

        # The following setters needs to be called after the UI has been built
        # as they need to modify some UI widgets.
        self.show_country_servers = show_country_servers
        self._connection_state = connection_state

    def _build_ui(self, connection_state: ConnectionStateEnum):
        self._country_name_label = Gtk.Label(label=self.country_name)
        self._country_name_label.set_halign(Gtk.Align.START)
        self.prepend(self._country_name_label)
        self.set_spacing(10)

        self._show_under_maintenance_icon_or_country_details()

        self._toggle_button = Gtk.Button()
        self._toggle_button.add_css_class("secondary")
        self._toggle_button_handler_id = self._toggle_button.connect(
            "clicked", self._on_toggle_button_clicked
        )
        self._country_name_label.set_halign(Gtk.Align.END)
        self.append(self._toggle_button)

        self.connection_state = connection_state

    def update_under_maintenance_status(self, under_maintenance: bool):
        """Shows or hides the under maintenance status for the country."""
        self._under_maintenance = under_maintenance
        self._show_under_maintenance_icon_or_country_details()

    def _show_under_maintenance_icon_or_country_details(self):
        if self._under_maintenance:
            self._show_under_maintenance_icon()
        else:
            self._show_country_details()

    def _show_under_maintenance_icon(self):
        if self._country_details:
            self._country_details.set_visible(False)

        if not self._under_maintenance_icon:
            self._under_maintenance_icon = UnderMaintenanceIcon(self.country_name)
            self._under_maintenance_icon.set_halign(Gtk.Align.END)
            self.append(self._under_maintenance_icon)

        self._country_name_label.set_property("sensitive", False)

    def _show_country_details(self):
        if self._under_maintenance_icon:
            self._under_maintenance_icon.set_visible(False)

        if not self._country_details:
            self._country_details = self._build_country_details()
            self.append(self._country_details)

        self._country_details.set_visible(True)
        self._country_name_label.set_property("sensitive", True)

    def _build_country_details(self):
        country_details = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        country_details.set_halign(Gtk.Align.END)
        country_details.set_hexpand(True)
        country_details.set_spacing(10)

        if self._upgrade_required:
            button = self._build_upgrade_required_link_button()
            country_details.append(button)
        else:
            button = self._build_connect_button()
            self._connect_button = button
            country_details.append(self._connect_button)

        add_accessibility(button, Gtk.AccessibleRelation.LABELLED_BY, self._country_name_label)

        country_row_icons = []
        if self._smart_routing:
            country_row_icons.append(SmartRoutingIcon())

        server_feature_icons = self._build_server_feature_icons()
        country_row_icons.extend(server_feature_icons)
        for icon in country_row_icons:
            country_details.prepend(icon)

        if country_row_icons:
            add_accessibility(button, Gtk.AccessibleRelation.DESCRIBED_BY, country_row_icons)

        return country_details

    @property
    def under_maintenance(self) -> bool:
        """Indicates whether all the servers for this country are under maintenance or not."""
        return self._under_maintenance

    @property
    def upgrade_required(self):
        """Indicates whether the user needs to upgrade to have access to this country or not."""
        return self._upgrade_required

    def _build_upgrade_required_link_button(self) -> Gtk.LinkButton:
        upgrade_button = Gtk.LinkButton.new_with_label("Upgrade")
        upgrade_button.set_uri("https://account.protonvpn.com/")
        return upgrade_button

    def _build_connect_button(self) -> Gtk.Button:
        connect_button = Gtk.Button()
        self._connect_button_handler_id = connect_button.connect(
            "clicked", self._on_connect_button_clicked
        )
        connect_button.add_css_class("secondary")
        return connect_button

    def _build_server_feature_icons(self) -> List[Gtk.Image]:
        server_feature_icons: List[Gtk.Image] = []
        if ServerFeatureEnum.P2P in self._server_features:
            server_feature_icons.append(P2PIcon())
        if ServerFeatureEnum.TOR in self._server_features:
            server_feature_icons.append(TORIcon())
        return server_feature_icons

    @property
    def server_features(self) -> Set[ServerFeatureEnum]:
        """Returns the set of features supported by the servers in this country."""
        return self._server_features

    @GObject.Signal(name="toggle-country-servers")
    def toggle_country_servers(self):
        """Signal when the user clicks the button to expand/collapse the servers
        from a country."""

    @property
    def country_code(self):
        """Returns the code of the country this header is for."""
        return self._country.code

    @property
    def country_name(self):
        """Returns the name of the country this header is for."""
        return self._country.name

    @property
    def show_country_servers(self):
        """Returns whether the country servers should be shown or not."""
        return self._show_country_servers

    @show_country_servers.setter
    def show_country_servers(self, show_country_servers: bool):
        """Sets whether the country servers should be shown or not."""
        self._show_country_servers = show_country_servers
        self._toggle_button.set_child(
            self._expanded_img if self.show_country_servers else self._collapsed_img
        )
        self._toggle_button.set_tooltip_text(
            f"Hide all servers from {self.country_name}" if self.show_country_servers else
            f"Show all servers from {self.country_name}"
        )

    @property
    def available(self) -> bool:
        """Returns True if the country is available, meaning the user can
        connect to one of its servers. Otherwise, it returns False."""
        return not self.upgrade_required and not self.under_maintenance

    @property
    def connection_state(self):
        """Returns the connection state of the server shown in this row."""
        return self._connection_state

    @connection_state.setter
    def connection_state(self, connection_state: ConnectionStateEnum):
        """Sets the connection state, modifying the row depending on the state."""
        # pylint: disable=duplicate-code
        self._connection_state = connection_state

        if self.available:
            # Update the server row according to the connection state.
            method = f"_on_connection_state_{connection_state.name.lower()}"
            if hasattr(self, method):
                getattr(self, method)()

    def _on_toggle_button_clicked(self, _toggle_button: Gtk.Button):
        self.show_country_servers = not self.show_country_servers
        self.emit("toggle-country-servers")

    def _on_connect_button_clicked(self, _connect_button: Gtk.Button):
        future = self._controller.connect_to_country(self.country_code)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.

    def _on_connection_state_disconnected(self):
        """Flags this server as "not connected"."""
        self._connect_button.set_sensitive(True)
        self._connect_button.set_label("Connect")

    def _on_connection_state_connecting(self):
        """Flags this server as "connecting"."""
        self._connect_button.set_label("Connecting...")
        self._connect_button.set_sensitive(False)

    def _on_connection_state_connected(self):
        """Flags this server as "connected"."""
        self._connect_button.set_sensitive(False)
        self._connect_button.set_label("Connected")

    def _on_connection_state_disconnecting(self):
        pass

    def _on_connection_state_error(self):
        """Flags this server as "error"."""
        self._on_connection_state_disconnected()

    def click_toggle_country_servers_button(self):
        """Clicks the button to toggle the country servers.
        This method was made available for tests."""
        self._toggle_button.emit("clicked")

    def click_connect_button(self):
        """Clicks the button to connect to the country.
        This method was made available for tests."""
        self._connect_button.emit("clicked")

    def grab_focus(self):  # pylint: disable=arguments-differ
        """Focuses on the connect button if available, otherwise on the toggle button."""
        if self._connect_button and not self._under_maintenance:
            self._connect_button.grab_focus()
        elif self._toggle_button:
            self._toggle_button.grab_focus()

    def cleanup(self):
        """Clean up signal connections to allow garbage collection."""
        if self._toggle_button_handler_id:
            self._toggle_button.disconnect(self._toggle_button_handler_id)
            self._toggle_button_handler_id = None
            self._toggle_button = None

        if self._connect_button_handler_id:
            self._connect_button.disconnect(self._connect_button_handler_id)
            self._connect_button_handler_id = None
            self._connect_button = None


class DeferredCountryRow(Gtk.Box):  # pylint: disable=too-many-instance-attributes
    """Row containing all servers, servers are loaded lazily"""

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            country: Country,
            user_tier: int,
            controller: Controller,
            connected_server_id: Optional[str] = None,
            show_country_servers: bool = False,
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self._controller = controller
        self._indexed_server_rows: dict[str, ServerRow] = {}

        free_servers, plus_servers =\
            self._group_servers_by_tier(country.servers)
        is_free_user = user_tier == 0

        # Properties initialized after building all server rows.
        self._is_free_country = None
        self._upgrade_required = None
        self._country_features = set()
        self._under_maintenance = None

        self._server_rows_revealer = Gtk.Revealer()
        server_rows_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        server_rows_container.set_spacing(10)
        self._server_rows_revealer.set_margin_top(15)
        self._server_rows_revealer.set_child(server_rows_container)

        ordered_servers = []
        if is_free_user:
            ordered_servers.extend(free_servers)
            ordered_servers.extend(plus_servers)
        else:
            ordered_servers.extend(plus_servers)
            ordered_servers.extend(free_servers)

        analysis = _analyze_servers(ordered_servers, connected_server_id)

        self._under_maintenance = analysis.under_maintenance  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
        self._is_free_country = analysis.is_free_country  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
        self._country_features = analysis.country_features  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
        self._connected_server_id = connected_server_id

        def add_servers_to_country():
            for server in ordered_servers:
                server_row = ServerRow(
                    server=server,
                    user_tier=user_tier,
                    controller=self._controller
                )
                server_rows_container.append(server_row)

                self._indexed_server_rows[server.id] = server_row

                # If we are currently connected to a server then set its row
                # state to "connected".
                #
                # We use self._connected_server_id instead of
                # connected_server_id because there's a chance it might change
                # before this function is called.
                if self._connected_server_id == server.id:
                    server_row.connection_state = ConnectionStateEnum.CONNECTED

        self._add_servers_to_country: Optional[Callable] = add_servers_to_country

        self._upgrade_required = is_free_user and not self._is_free_country

        self._country_header = CountryHeader(
            country=country,
            under_maintenance=self._under_maintenance,
            upgrade_required=self._upgrade_required,
            server_features=self._country_features,
            smart_routing=analysis.smart_routing_country,
            connection_state=analysis.country_connection_state,
            controller=controller,
            show_country_servers=show_country_servers
        )
        self._toggle_signal_handler_id = self._country_header.connect(
            "toggle-country-servers", self._on_toggle_country_servers
        )

        self.append(self._country_header)
        self.append(self._server_rows_revealer)

        if show_country_servers:
            self._server_rows_revealer.set_reveal_child(True)

    def _generate_servers_if_needed(self, country_header: CountryHeader):
        if country_header.show_country_servers:
            if self._add_servers_to_country:
                self._add_servers_to_country()
                self._add_servers_to_country = None

    def toggle_row(self):
        """Toggles the view of the children of the country row."""
        self._country_header.click_toggle_country_servers_button()

    @property
    def country_code(self):
        """Returns the code of the country. A short unique string"""
        return self._country_header.country_code

    @property
    def country_name(self):
        """Returns the name of the country.
        This method was made available for tests."""
        return self._country_header.country_name

    @property
    def upgrade_required(self):
        """Returns True if this country is not in the currently logged-in
        user tier, and therefore it requires a plan upgrade. Otherwise, it
        returns False."""
        return self._upgrade_required

    @property
    def is_free_country(self) -> bool:
        """Returns True if this country has any servers available to
        users with a free account. Otherwise, it returns False."""
        return bool(self._is_free_country)

    @property
    def showing_servers(self):
        """Returns True if the servers are being showed and False otherwise.
        This method was made available for tests."""
        return self._server_rows_revealer.get_reveal_child()

    def click_toggle_country_servers_button(self):
        """
        Clicks the button to toggle the visibility of the country servers.
        This method was made available for tests.
        """
        self._country_header.click_toggle_country_servers_button()

    def grab_focus(self):  # pylint: disable=arguments-differ
        """Focuses on the country row."""
        self._country_header.grab_focus()

    @property
    def server_rows(self) -> List[ServerRow]:
        """Returns the list of server rows for this server.
        This method was made available for tests."""
        server_rows: List[ServerRow] = []
        revealer_child = self._server_rows_revealer.get_child()
        if revealer_child:
            child = revealer_child.get_first_child()
            while child:
                server_rows.append(child)
                child = child.get_next_sibling()
        return server_rows

    @property
    def connection_state(self):
        """Returns the connection state for this row."""
        return self._country_header.connection_state

    @property
    def header_searchable_content(self) -> str:
        """Returns the normalized searchable content for the country header."""
        return normalize(self.country_name)

    @staticmethod
    def _group_servers_by_tier(country_servers) -> Tuple[List[LogicalServer], List[LogicalServer]]:
        free_servers = []
        plus_servers = []
        for server in country_servers:
            if server.tier == 0:
                free_servers.append(server)
            else:
                plus_servers.append(server)

        return free_servers, plus_servers

    def _on_toggle_country_servers(self, country_header: CountryHeader):
        self._server_rows_revealer.set_reveal_child(
            country_header.show_country_servers
        )
        self._generate_servers_if_needed(country_header)

    def set_servers_visibility(self, visible: bool):
        """Country servers will be shown if set to True. Otherwise, they'll be hidden."""
        self._country_header.show_country_servers = visible
        self._server_rows_revealer.set_reveal_child(visible)

    def connection_status_update(self, connection_state):
        """This method is called by VPNWidget whenever the VPN connection status changes."""
        self._country_header.connection_state = connection_state.type
        server_id = connection_state.context.connection.server_id
        server = self._indexed_server_rows.get(server_id, None)
        if server:
            server.connection_state = connection_state.type

        # maintain connected server id only when connected
        if self._controller.is_connection_active:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            self._connected_server_id = server_id
        else:
            self._connected_server_id = None

    def click_connect_button(self):
        """Clicks the button to connect to the country.
        This method was made available for tests."""
        self._country_header.click_connect_button()

    def update_server_loads(self, new_country: Country):
        """Refreshes the UI after new server loads were retrieved."""
        # Start by setting the country under maintenance until the opposite is proven.
        for server_row in self._indexed_server_rows.values():
            server_row.update_server_load()

        if new_country is not None:
            self._under_maintenance = _analyze_servers(
                new_country.servers,
                self._connected_server_id
            ).under_maintenance

            self._country_header.update_under_maintenance_status(
                self._under_maintenance
            )

    def cleanup(self):
        """Clean up signal connections and references to allow garbage collection."""
        # Disconnect country header signal
        self._country_header.disconnect(self._toggle_signal_handler_id)

        # Clean up country header's signal connections
        self._country_header.cleanup()
        self._country_header = None

        # Clean up all server rows and their signal connections
        for server_row in self._indexed_server_rows.values():
            server_row.cleanup()
        self._indexed_server_rows.clear()
        self._indexed_server_rows = {}

        # Clear the deferred server creation function and its captured variables
        self._add_servers_to_country = None

        # Clear other references
        # Remove child to break widget tree references
        child = self._server_rows_revealer.get_child()
        if child:
            self._server_rows_revealer.set_child(None)
        self._server_rows_revealer = None

        # Clear controller reference
        self._controller = None



"""
This module defines the connection status widget.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Optional, cast
from gi.repository import GLib
from proton.vpn.app.gtk import Gtk
from proton.vpn.connection import events, states
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget, LoadingConnectionWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.vpn.port_forward_widget import PortForwardRevealer
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.split_tunneling import \
    SPLIT_TUNNELING_TOGGLE_SETTING_NAME
from proton.vpn import logging

logger = logging.getLogger(__name__)

SPLIT_TUNNELING_APP_RESTART_MESSAGE = \
    "Split tunneling enabled. Remember to restart affected apps."


class VPNConnectionStatusWidget(Gtk.Box):
    """Displays the current connection status."""
    MAXIMUM_SESSIONS_ERROR = "You've reached your maximum device limit. " \
        "To reconnect to VPN, please disconnect from another device."

    def __init__(
        self, controller: Controller,
        overlay_widget: OverlayWidget,
        notifications: Notifications,
        port_forward_revealer: Optional[PortForwardRevealer] = None
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self.set_name("vpn-connection-status-widget")
        self._overlay_widget = overlay_widget
        self._controller = controller
        self._notifications = notifications

        self._connection_status_label = Gtk.Label(label="")
        self._connection_status_label.set_name("connection-status-label")
        self._connection_status_label.set_wrap(True)
        # Setting max_width_chars to a small value prevents the label from
        # requesting extra horizontal space, forcing it to wrap within its
        # allocated width instead.
        self._connection_status_label.set_max_width_chars(1)
        self._loading_widget = self._build_loading_connection_widget()

        self.append(self._connection_status_label)

        self._port_forward_revealer: Optional[PortForwardRevealer] = None
        display_port_forwarding = controller.feature_flags\
            .get("DisplayPortForwarding")
        if display_port_forwarding:
            self._port_forward_revealer = port_forward_revealer \
                or PortForwardRevealer(notifications)
            self.append(self._port_forward_revealer)

    def _build_loading_connection_widget(self) -> LoadingConnectionWidget:
        cancel_button = Gtk.Button.new_with_label("Cancel Connection")
        cancel_button.connect("clicked", self._on_cancel_button_clicked)

        loading_widget = LoadingConnectionWidget(
            label="",
            cancel_button=cancel_button
        )

        return loading_widget

    def _on_cancel_button_clicked(self, _):
        logger.info("Disconnect from VPN", category="ui", event="disconnect")
        future = self._controller.disconnect()
        future.add_done_callback(lambda f: GLib.idle_add(f.result))

    @property
    def status_message(self) -> str:
        """Returns the connection status message being displayed to the user."""
        return self._connection_status_label.get_label()

    def connection_status_update(self, connection_state: states.State):
        """This method is called by VPNWidget whenever the VPN connection status changes."""
        self._update_connection_status_label(connection_state)

    def _update_connection_status_label(self, connection_state: states.State):
        connection = connection_state.context.connection

        label = ""
        if isinstance(connection_state, states.Disconnected):
            label = "You are disconnected"
            self._overlay_widget.hide()
        elif isinstance(connection_state, states.Connecting):
            self._loading_widget.set_label(f"Connecting to {connection.server_name}")
            self._overlay_widget.show(self._loading_widget)
        elif isinstance(connection_state, states.Connected):
            label = f"You are connected to {connection.server_name}"
            self._overlay_widget.hide()
            if self._split_tunneling_enabled:
                self._notifications.show_info_message(
                    message=SPLIT_TUNNELING_APP_RESTART_MESSAGE
                )
        elif isinstance(connection_state, states.Disconnecting):
            label = f"Disconnecting from {connection.server_name}"
        elif isinstance(connection_state, states.Error):
            last_connection_event = connection_state.context.event
            label = "Connection error"
            if isinstance(last_connection_event, events.TunnelSetupFailed):
                label = f"{label}: tunnel setup failed"
            elif isinstance(last_connection_event, events.AuthDenied):
                label = f"{label}: authentication denied"
            elif isinstance(last_connection_event, events.Timeout):
                label = f"{label}: timeout"
            elif isinstance(last_connection_event, events.DeviceDisconnected):
                label = f"{label}: device disconnected"
            elif isinstance(last_connection_event, events.MaximumSessionsReached):
                label = f"{label}: session limit reached"
                self._notifications.show_error_dialog(
                    message=self.MAXIMUM_SESSIONS_ERROR,
                    title=label
                )

            self._overlay_widget.hide()

        # This condition will be removed once we remove the feature flag.
        if self._port_forward_revealer:
            self._port_forward_revealer.on_new_state(connection_state)

        self._connection_status_label.set_label(label)

    @property
    def _split_tunneling_enabled(self) -> bool:
        """Check if split tunneling is enabled."""
        return cast(bool, self._controller.get_setting_attr(SPLIT_TUNNELING_TOGGLE_SETTING_NAME))



"""
Connection settings module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING

from proton.vpn.app.gtk.conflicts import WIREGUARD_PROTOCOL

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    BaseCategoryContainer, ToggleWidget, ConflictableComboboxWidget,
    ReactiveSettingContainer, ReactiveSetting
)
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.custom_dns import CustomDNSWidget

if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.headerbar.menu.settings.settings_window import \
        SettingsWindow


class ProtocolComboboxWidget(ConflictableComboboxWidget, ReactiveSetting):
    """Combobox widget for selecting the VPN protocol."""
    def on_settings_changed(self, settings):
        if self.combobox.get_active_text() != settings.protocol:
            self.combobox.set_active_id(settings.protocol)


class ConnectionSettings(BaseCategoryContainer, ReactiveSettingContainer):  # noqa: E501 # pylint: disable=line-too-long, too-many-instance-attributes
    """Settings related to connection are all grouped under this class."""
    CATEGORY_NAME = "Connection"
    PROTOCOL_LABEL = "Protocol"
    PROTOCOL_DESCRIPTION = "Protocol can only be changed when VPN is disconnected."
    VPN_ACCELERATOR_LABEL = "VPN Accelerator"
    VPN_ACCELERATOR_DESCRIPTION = "Increase your connection speed by up to 400% "\
        "with performance enhancing technologies."
    MODERATE_NAT_LABEL = "Moderate NAT"
    MODERATE_NAT_DESCRIPTION = "Disables randomization of the local addresses mapping. "\
        "This can slightly reduce connection security, but should allow direct "\
        "connections for online gaming and similar purposes."
    SWITCH_PROTOCOL_IF_CONNECTION_ACTIVE_DESCRIPTION = "Protocol selection "\
        "is disabled while VPN is active. Disconnect to make changes."
    IPV6_LABEL = "IPv6"
    IPV6_DESCRIPTION = "Tunnels IPv6 traffic through the VPN. "\
        "Can enhance compatibility with IPv6 networks."

    def __init__(self, controller: Controller, settings_window: "SettingsWindow"):
        super().__init__(self.CATEGORY_NAME)
        self._controller = controller
        self._settings_window = settings_window
        self.custom_dns = None

    def build_ui(self):
        """Builds the UI, invoking all necessary methods that are
        under this category."""
        self.build_protocol()
        self.build_vpn_accelerator()
        self.build_moderate_nat()
        self.build_ipv6()
        self.build_custom_dns()

    def build_protocol(self):
        """Builds and adds the `protocol` setting to the widget."""
        protocol_list_of_tuples = [
            (protocol.cls.protocol, protocol.cls.ui_protocol)
            for protocol in self._controller.get_available_protocols()
        ]

        def do_set(combobox, new_value: str):
            combobox.save_setting(new_value)

        def do_revert(combobox):
            combobox.combobox.set_active_id(WIREGUARD_PROTOCOL)

        self.append(ProtocolComboboxWidget(
                controller=self._controller,
                title=self.PROTOCOL_LABEL,
                description=self.PROTOCOL_DESCRIPTION,
                setting_name="settings.protocol",
                combobox_options=protocol_list_of_tuples,
                disable_on_active_connection=True,
                do_set=do_set,
                do_revert=do_revert
            ))

    def build_vpn_accelerator(self):
        """Builds and adds the `vpn_accelerator` setting to the widget."""
        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
            toggle_widget.save_setting(new_value)

        self.append(ToggleWidget(
            controller=self._controller,
            title=self.VPN_ACCELERATOR_LABEL,
            description=self.VPN_ACCELERATOR_DESCRIPTION,
            setting_name="settings.features.vpn_accelerator",
            requires_subscription_to_be_active=True,
            callback=on_switch_state
        ))

    def build_moderate_nat(self):
        """Builds and adds the `moderate_nat` setting to the widget."""
        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
            toggle_widget.save_setting(new_value)

        self.append(ToggleWidget(
            controller=self._controller,
            title=self.MODERATE_NAT_LABEL,
            description=self.MODERATE_NAT_DESCRIPTION,
            setting_name="settings.features.moderate_nat",
            requires_subscription_to_be_active=True,
            callback=on_switch_state
        ))

    def build_ipv6(self):
        """Builds and adds the `ipv6` setting to the widget."""
        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
            toggle_widget.save_setting(new_value)
            self._settings_window.notify_user_with_reconnect_message(force_notify=True)

        self.append(ToggleWidget(
            controller=self._controller,
            title=self.IPV6_LABEL,
            description=self.IPV6_DESCRIPTION,
            setting_name="settings.ipv6",
            callback=on_switch_state
        ))

    def build_custom_dns(self):
        """Builds and adds the `custom_dns` setting to the widget."""
        self.custom_dns = CustomDNSWidget.build(self._controller, self._settings_window)
        self.append(self.custom_dns)



"""
This module defines the widgets used to present the VPN server list to the user.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import List, Dict, Optional

from gi.repository import GLib, GObject

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.vpn.serverlist.country import DeferredCountryRow
from proton.vpn.app.gtk.widgets.vpn.search_entry import SearchEntry
from proton.vpn.session.servers import Country, LogicalServer, ServerList
from proton.vpn import logging


logger = logging.getLogger(__name__)


@dataclass
class ServerListWidgetState:
    """
    Holds the state of the ServerListWidget. This state is reset after
    login/logout.

    Attributes:
        user_tier: the tier the user has access to.
        server_list: list of servers to be displayed.
        country_rows: country rows indexed by country code.
    """
    user_tier: Optional[int] = None
    server_list: Optional[ServerList] = None
    country_rows: Dict[str, DeferredCountryRow] = field(default_factory=dict)

    def get_server_by_id(self, server_id: str) -> LogicalServer:
        """Returns the server with the given name."""
        if self.server_list:
            return self.server_list.get_by_id(server_id)
        return None


class ServerListWidget(Gtk.ScrolledWindow):
    """Displays the VPN servers list."""

    # Number of seconds to wait before checking if the servers cache expired.
    RELOAD_INTERVAL_IN_SECONDS = 60

    def __init__(self, controller: Controller, search_entry: SearchEntry | None = None):
        super().__init__()
        self._search_entry = search_entry
        self.set_name("server-list-widget")
        self.set_policy(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC
        )
        self.set_propagate_natural_width(True)
        self._controller = controller
        self._container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._container.set_name("server-list-widget-container")
        self._container.set_vexpand(True)
        self._container.set_margin_end(15)  # Leave space for the scroll bar.
        self._container.set_spacing(5)
        self.set_child(self._container)

        self._state = ServerListWidgetState()

        self.connect("unrealize", self._on_unrealize)

    def _on_unrealize(self, _widget):
        self.unload()

    @GObject.Signal(name="filter-complete")
    def filter_complete(self):
        """Signal emitted after the UI finalized filtering the UI."""

    @GObject.Signal(name="ui-updated")
    def ui_updated(self):
        """Signal emitted once the server list within the UI has been updated.
        Mainly used for test purposes."""

    @property
    def country_rows(self) -> List[DeferredCountryRow]:
        """Returns the list of country rows that are currently being displayed.
        This method was made available for tests."""
        return list(self._state.country_rows.values())

    def connection_status_update(self, connection_status):
        """
        This method is called by VPNWidget whenever the VPN connection status changes.
        Important: as this method is always called from another thread, we need
        to make sure that any resulting actions are passed to the main thread
        running GLib's main loop with GLib.idle_add.
        """
        connection = connection_status.context.connection
        if connection:
            def update_server_rows():
                country_row = self._get_country_row(connection.server_id)
                country_row.connection_status_update(connection_status)

            GLib.idle_add(update_server_rows)

    def _remove_country_rows(self):
        """Remove UI country rows."""
        row = self._container.get_first_child()
        while row:
            next_row = row.get_next_sibling()  # Get next before removing
            # Clean up signal connections and references before removing
            row.cleanup()
            self._container.remove(row)
            row = next_row

    def _on_server_list_update(self):
        """Whenever a new server list is received the UI should be updated."""
        start = time.time()
        self._state.server_list = self._controller.server_list
        self._build_country_rows()
        logger.info(
            "Full server list widget update completed in "
            f"{time.time() - start:.2f} seconds."
        )

    def _on_server_loads_update(self):
        start = time.time()

        new_countries = {
            country.code: country
            for country in self._controller.server_list.group_by_country()
        }

        for country_row in self._state.country_rows.values():
            new_country = new_countries.get(country_row.country_code, None)
            country_row.update_server_loads(new_country)

        logger.info(
            "Partial server list widget update completed in "
            f"{time.time() - start:.2f} seconds."
        )

    def focus_on_entry(self, _widget, name_to_search: str) -> None:
        """Searches for an entry by name and either connects to it directly,
           or focuses on it."""
        for country in self.country_rows:

            # Server
            if "#" in name_to_search:
                future = self._controller.connect_to_server(name_to_search)
                future.add_done_callback(lambda f: GLib.idle_add(f.result))
                if self._search_entry:
                    self._search_entry.grab_focus()
                return

            # Country
            if country.country_name.lower() == name_to_search.lower():
                if not country.showing_servers:
                    country.toggle_row()
                country.grab_focus()
                return

    def display(self, user_tier: int, server_list: int):
        """Update UI with the new server list."""
        # Create new state (old state cleanup is handled in _build_country_rows)
        self._state = ServerListWidgetState(
            server_list=server_list,
            user_tier=user_tier
        )

        self._build_country_rows()

        self._controller.set_server_list_updated_callback(self._on_server_list_update)
        self._controller.set_server_loads_updated_callback(self._on_server_loads_update)

    def _build_country_rows(self):
        # Create new country rows
        self._state.country_rows = self._create_new_country_rows(
            old_country_rows=self._state.country_rows
        )
        # Remove old rows from UI and clean them up only after new rows are added
        # to state since the old rows are still needed for the new rows to be created.
        self._remove_country_rows()

        self._add_country_rows()
        self.emit("ui-updated")

    def unload(self):
        """Things to do before the widget is being removed from the window."""
        self._controller.unset_server_list_updated_callback()
        self._controller.unset_server_loads_updated_callback()

    def _add_country_rows(self):
        """Adds country rows to the container."""
        for country_row in self._state.country_rows.values():
            self._container.append(country_row)

    def _create_new_country_rows(self, old_country_rows) -> Dict[str, DeferredCountryRow]:
        """Returns new country rows."""
        countries = self._state.server_list.group_by_country()
        if self._state.user_tier == 0:
            # If the current user has a free account, sort the countries having
            # free servers first.
            countries.sort(key=free_countries_first_sorting_key)

        connected_server_id = None
        if self._controller.is_connection_active:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            connected_server_id = self._controller.current_server_id

        new_country_rows = {}
        for country in countries:
            show_country_servers = False
            if old_country_rows and old_country_rows.get(country.code):
                show_country_servers = old_country_rows[country.code].showing_servers

            country_row = DeferredCountryRow(
                country=country,
                user_tier=self._state.user_tier,
                controller=self._controller,
                connected_server_id=connected_server_id,
                show_country_servers=show_country_servers
            )
            new_country_rows[country.code.lower()] = country_row

        return new_country_rows

    def _get_country_row(self, server_id: str) -> DeferredCountryRow:
        """Returns a country row based on the vpn server."""
        logical_server = self._state.get_server_by_id(server_id)
        country_code = logical_server.exit_country.lower()
        try:
            return self._state.country_rows[country_code]
        except KeyError as error:
            raise RuntimeError(
                f"Unable to get country row {country_code} for server "
                f"{server_id}."
            ) from error


def free_countries_first_sorting_key(country: Country):
    """
    Returns the comparison key to sort countries according to
    business rules for free users.

    Apart from sorting country rows by country name, free users should
    have countries having free servers sorted first.

    :param country: country row to generate the comparison key for.
    :return: The comparison key.
    """
    return (0 if country.free else 1, country.name)  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses



"""
General settings module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING, Optional
from gi.repository import Gtk
from proton.vpn import logging
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    BaseCategoryContainer, ToggleWidget, EntryWidget,
)
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.early_access import \
    EarlyAccessWidget
if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.main.tray_indicator import TrayIndicator

logger = logging.getLogger(__name__)


class TrayPinnedServersWidget(EntryWidget):
    """Custom widget that holds the pinned servers to tray setting."""
    TRAY_PINNED_SERVERS_LABEL = "Pinned tray connections"
    TRAY_PINNED_SERVERS_DESCRIPTION = "Access preferred connections from system tray."\
        " Enter country or server codes, separated by commas, to quickly connect "\
        "(e.g.: NL#42, JP, US, IT#01)."
    SETTING_NAME = "app_configuration.tray_pinned_servers"

    def __init__(self, controller: Controller, tray_indicator: Optional["TrayIndicator"] = None):
        super().__init__(
            controller=controller,
            title=self.TRAY_PINNED_SERVERS_LABEL,
            description=self.TRAY_PINNED_SERVERS_DESCRIPTION,
            setting_name=self.SETTING_NAME,
            callback=self._on_focus_outside_entry
        )
        self._controller = controller
        self._tray_indicator = tray_indicator

    def _on_focus_outside_entry(self, entry: Gtk.Entry, *_):
        self.save_setting(entry.get_text())
        if self._tray_indicator:
            self._tray_indicator.reload_pinned_servers()

    def get_setting(self):
        """Shortcut property that sets the new setting and stores to disk."""
        tray_pinned_servers = self._controller.get_setting_attr(self.SETTING_NAME)
        return ', '.join(tray_pinned_servers)

    def save_setting(self, new_value: str):  # noqa: F811
        """Returns if the the upgrade tag has overridden original interactive
        object."""
        server_list = []

        for pinned_server in new_value.split(","):
            cleaned_pinned_server = pinned_server.strip().upper()

            if cleaned_pinned_server:
                server_list.append(cleaned_pinned_server)

        self._controller.save_setting_attr(self.SETTING_NAME, server_list)


class GeneralSettings(BaseCategoryContainer):  # pylint: disable=too-many-instance-attributes
    """General settings are grouped under this class."""
    CATEGORY_NAME = "General"
    CONNECT_AT_APP_STARTUP_LABEL = "Auto connect"
    CONNECT_AT_APP_STARTUP_DESCRIPTION = "You will be connected to a server as "\
        "soon as Proton VPN app starts. Replace it with a country ISO code "\
        "(e.g.: US for United States), a server (e.g.: NL#42)"\
        " or Fastest for quick connection. Default value: Off."
    START_APP_MINIMIZED_LABEL = "Start app minimized"
    START_APP_MINIMIZED_DESCRIPTION = "When enabled, the app starts minimized "\
        "to the tray."
    ANONYMOUS_CRASH_REPORTS_LABEL = "Share anonymous crash reports"
    ANONYMOUS_CRASH_REPORTS_DESCRIPTION = "Crash reports help us fix bugs, detect firewalls, "\
        "and avoid VPN blocks.\n\nThese statistics do not contain your IP address, and they "\
        "cannot be used to identify you. We'll never share them with third parties."

    def __init__(
        self, controller: Controller,
        tray_indicator: Optional["TrayIndicator"] = None,
    ):
        super().__init__(self.CATEGORY_NAME)
        self._controller = controller
        self._tray_indicator = tray_indicator

    def build_ui(self):
        """Builds the UI, invoking all necessary methods that are
        under this category."""
        self.build_connect_at_app_startup()

        if self._tray_indicator:
            self.build_start_app_minimized()
            self.build_tray_pinned_servers()

        self.build_anonymous_crash_reports()
        self.build_beta_upgrade()

    def build_connect_at_app_startup(self):
        """Builds and adds the `connect_at_app_startup` setting to the widget."""
        def on_focus_out_callback(entry: Gtk.Entry, entry_widget: EntryWidget, *_):
            new_value: Optional[str] = entry.get_text().strip().upper()
            if new_value == "OFF":
                new_value = None

            entry_widget.save_setting(new_value)

        self.append(EntryWidget(
            controller=self._controller,
            title=self.CONNECT_AT_APP_STARTUP_LABEL,
            description=self.CONNECT_AT_APP_STARTUP_DESCRIPTION,
            setting_name="app_configuration.connect_at_app_startup",
            callback=on_focus_out_callback
        ))

    def build_start_app_minimized(self):
        """Builds and adds the `start_app_minimized` setting to the widget."""
        self.append(ToggleWidget(
            controller=self._controller,
            title=self.START_APP_MINIMIZED_LABEL,
            description=self.START_APP_MINIMIZED_DESCRIPTION,
            setting_name="app_configuration.start_app_minimized"
        ))

    def build_tray_pinned_servers(self):
        """Builds and adds the `tray_pinned_servers` setting to the widget."""
        self.append(TrayPinnedServersWidget(
            controller=self._controller, tray_indicator=self._tray_indicator
        ))

    def build_anonymous_crash_reports(self):
        """Builds and adds the `anonymous_crash_reports` setting to the widget."""
        self.append(ToggleWidget(
            controller=self._controller,
            title=self.ANONYMOUS_CRASH_REPORTS_LABEL,
            description=self.ANONYMOUS_CRASH_REPORTS_DESCRIPTION,
            setting_name="settings.anonymous_crash_reports"
        ))

    def build_beta_upgrade(self):
        """Builds and adds the `Early Access` setting to the widget."""
        early_access = EarlyAccessWidget(self._controller)

        if not early_access.can_early_access_be_displayed():
            return

        self.append(early_access)



"""
This module defines the server widget.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List, Optional

from gi.repository import GLib, Pango

from proton.vpn.app.gtk.utils.accessibility import add_accessibility
from proton.vpn.app.gtk.utils.search import normalize
from proton.vpn.connection.enum import ConnectionStateEnum
from proton.vpn.session.servers import LogicalServer, ServerFeatureEnum
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import \
    UnderMaintenanceIcon, SmartRoutingIcon, StreamingIcon, \
    P2PIcon, TORIcon, SecureCoreIcon
from proton.vpn.app.gtk import Gtk
from proton.vpn import logging

from proton.vpn.app.gtk.controller import Controller

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class ServerRow(Gtk.Box):
    """Displays a single server as a row in the server list."""
    def __init__(self, server: LogicalServer, user_tier: int, controller: Controller):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL)
        self._server = server
        self._user_tier = user_tier
        self._controller = controller
        self._connection_state: ConnectionStateEnum = None
        self._server_details: Optional[Gtk.Box] = None
        self._icons_displayed = []
        self._under_maintenance_icon: Optional[UnderMaintenanceIcon] = None
        self._server_load: Optional[ServerLoad] = None
        self._connect_button: Optional[Gtk.Button] = None
        self._connect_button_handler_id: Optional[int] = None

        self._build_row()

    @property
    def connection_state(self):
        """Returns the connection state of the server shown in this row."""
        return self._connection_state

    @connection_state.setter
    def connection_state(self, connection_state: ConnectionStateEnum):
        """Sets the connection state, modifying the row depending on the state."""
        self._connection_state = connection_state

        if (
            self._connection_state == ConnectionStateEnum.CONNECTED
            and not self.available
        ):
            logger.warning(
                "Received connected state but server is not available",
                category="ui", event="conn:state"
            )

        if self.available:
            # Update the server row according to the connection state.
            method = f"_on_connection_state_{connection_state.name.lower()}"
            if hasattr(self, method):
                getattr(self, method)()

    def _build_row(self):
        self._server_label = Gtk.Label(label=self._server.name)
        # Some test server names are very long.
        self._server_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.append(self._server_label)

        self._show_under_maintenance_icon_or_server_details(self._server.enabled)

    def _show_under_maintenance_icon_or_server_details(self, server_enabled: bool):
        if server_enabled:
            self._show_server_details()
        else:
            self._show_under_maintenance_icon()

    def _show_under_maintenance_icon(self):
        if self._server_details:
            self._server_details.set_visible(False)

        if not self._under_maintenance_icon:
            self._under_maintenance_icon = UnderMaintenanceIcon(self._server.name)
            self.append(self._under_maintenance_icon)

        self._under_maintenance_icon.set_visible(True)
        self._server_label.set_property("sensitive", False)

    def _show_server_details(self):
        if self._under_maintenance_icon:
            self._under_maintenance_icon.set_visible(False)

        if not self._server_details:
            self._server_details = self._build_server_details()
            self.append(self._server_details)

        self._server_details.set_visible(True)
        self._server_label.set_property("sensitive", True)

    def _build_server_details(self) -> Gtk.Box:
        server_details = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        server_details.set_halign(Gtk.Align.END)
        server_details.set_hexpand(True)
        server_details.set_spacing(10)

        server_row_icons = []

        # If server supports Secure Core then it should be the only
        # icon to be displayed.
        if ServerFeatureEnum.SECURE_CORE in self._server.features:
            server_row_icons.append(
                SecureCoreIcon(self._server.entry_country_name, self._server.exit_country_name)
            )
        else:
            smart_routing = self._server.host_country is not None
            if smart_routing:
                server_row_icons.append(SmartRoutingIcon())

            server_feature_icons = self._build_server_feature_icons()
            server_row_icons.extend(server_feature_icons)

        for icon in server_row_icons:
            server_details.append(icon)
            self._icons_displayed.append(icon)

        self._server_load = ServerLoad(self._server.load)
        server_details.append(self._server_load)

        if self.upgrade_required:
            button = self._build_upgrade_link_button()
            server_details.append(button)
        else:
            self._connect_button = self._build_connect_button()
            button = self._connect_button
            server_details.append(self._connect_button)

        add_accessibility(button, Gtk.AccessibleRelation.LABELLED_BY, self._server_label)
        add_accessibility(
            button,
            Gtk.AccessibleRelation.DESCRIBED_BY,
            server_row_icons + [self._server_load]
        )

        return server_details

    def _build_connect_button(self):
        connect_button = Gtk.Button(label="Connect")
        self._connect_button_handler_id = connect_button.connect(
            "clicked", self._on_connect_button_clicked
        )
        connect_button.add_css_class("secondary")
        return connect_button

    def _build_upgrade_link_button(self):
        upgrade_button = Gtk.LinkButton.new_with_label("Upgrade")
        upgrade_button.set_uri("https://account.protonvpn.com/")
        return upgrade_button

    def _build_server_feature_icons(self) -> List[Gtk.Image]:
        server_feature_icons: List[Gtk.Image] = []

        if self._server.tier > 0:
            server_feature_icons.append(StreamingIcon())
        if ServerFeatureEnum.P2P in self._server.features:
            server_feature_icons.append(P2PIcon())
        if ServerFeatureEnum.TOR in self._server.features:
            server_feature_icons.append(TORIcon())

        return server_feature_icons

    def _on_connection_state_disconnected(self):
        """Flags this server as "not connected"."""
        self._connect_button.set_sensitive(True)
        self._connect_button.set_tooltip_text(f"Connect to {self.server_label}")
        self._connect_button.set_label("Connect")

    def _on_connection_state_connecting(self):
        """Flags this server as "connecting"."""
        self._connect_button.set_label("Connecting...")
        self._connect_button.set_tooltip_text(f"Connecting to {self.server_label}...")
        self._connect_button.set_sensitive(False)

    def _on_connection_state_connected(self):
        """Flags this server as "connected"."""
        self._connect_button.set_sensitive(False)
        self._connect_button.set_tooltip_text(f"Connected to {self.server_label}")
        self._connect_button.set_label("Connected")

    def _on_connection_state_disconnecting(self):
        pass

    def _on_connection_state_error(self):
        """Flags this server as "not connected"."""
        self._on_connection_state_disconnected()

    def _on_connect_button_clicked(self, _):
        future = self._controller.connect_to_server(self._server.name)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.

    @property
    def available(self) -> bool:
        """Returns True if the country is available, meaning the user can
        connect to one of its servers. Otherwise, it returns False."""
        return not self.upgrade_required and not self.under_maintenance

    @property
    def upgrade_required(self) -> bool:
        """Returns if a plan upgrade is required to connect to server."""
        return self._server.tier > self._user_tier

    @property
    def server_label(self) -> str:
        """Returns the server label."""
        return self._server_label.get_label()

    @property
    def server_id(self) -> str:
        """Returns the server ID."""
        return self._server.id

    @property
    def server_tier(self) -> int:
        """Returns the server tier."""
        return self._server.tier

    @property
    def under_maintenance(self) -> bool:
        """Returns if the server is under maintenance."""
        return not self._server.enabled

    @property
    def searchable_content(self) -> str:
        """Returns searchable content on this server."""
        return normalize(self.server_label)

    def click_connect_button(self):
        """Clicks the connect button.
        This method was made available for tests."""
        self._connect_button.emit("clicked")

    @property
    def is_connect_button_visible(self) -> bool:
        """Returns if the connect button is visible.
        This method was made available for tests."""
        return self._connect_button.is_visible()

    @property
    def server_load_label(self) -> str:
        """Returns the text shown as server load."""
        return self._server_load.get_text()

    @property
    def under_maintenance_icon_visible(self) -> bool:
        """Whether the under maintenance icon is shown or not."""
        return bool(self._under_maintenance_icon and self._under_maintenance_icon.is_visible())

    def is_server_feature_icon_displayed(self, icon_class):
        """Returns True if an instance of the specified icon class is displayed
        or False otherwise."""
        if not self._server_details.is_visible():
            return False

        filtered_icons = []
        child = self._server_details.get_first_child()
        while child:
            if isinstance(child, icon_class):
                filtered_icons.append(child)
            child = child.get_next_sibling()

        return bool(filtered_icons)

    def update_server_load(self):
        """Redraws the row after a server load update."""
        # The server status may have changed
        self._show_under_maintenance_icon_or_server_details(self._server.enabled)
        if self._server.enabled:
            self._server_load.set_load(self._server.load)

    def cleanup(self):
        """Clean up signal connections and references to allow garbage collection."""
        if self._connect_button_handler_id:
            self._connect_button.disconnect(self._connect_button_handler_id)
            self._connect_button_handler_id = None
            self._connect_button = None

        # Clear references
        self._controller = None
        self._server = None
        if self._icons_displayed:
            self._icons_displayed.clear()


class ServerLoad(Gtk.Label):
    """Displays the server load shown in a server row."""
    def __init__(self, load: int):
        super().__init__()
        self.set_name("server-load")

        self.set_load(load)

    def set_load(self, load: int):
        """Sets the load percentage to be displayed."""
        self.set_label(f"{load}%")
        help_text = f"Server load is at {load}%"
        self.set_tooltip_text(help_text)

        for cls in "signal-danger", "signal-warning", "signal-success":
            self.remove_css_class(cls)

        if load > 90:
            self.add_css_class("signal-danger")
        elif load > 75:
            self.add_css_class("signal-warning")
        else:
            self.add_css_class("signal-success")



"""
This module contains utility tools.

Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Callable
from gi.repository import Gtk

# See: https://docs.gtk.org/gtk4/migrating-3to4.html#set-a-proper-application-id  # pylint: disable=line-too-long # noqa: E501
APPLICATION_ID = "proton.vpn.app.gtk"


def connect_once(widget: Gtk.Widget, signal: str, callback: Callable, *args):
    """Subscribes to the signal once."""
    context = {}

    def wrapper(*args, **kwargs):
        widget.disconnect(context["id"])
        callback(*args, **kwargs)

    context["id"] = widget.connect(signal, wrapper, *args)
    return context["id"]


__all__ = ["connect_once"]
